'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const typeAnalyzer = require('@previewjs/type-analyzer');
const assertNever = require('assert-never');
const ts = require('typescript');
const prettier = require('prettier');
const parserBabel = require('prettier/parser-babel');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e["default"] : e; }

const assertNever__default = /*#__PURE__*/_interopDefaultLegacy(assertNever);
const ts__default = /*#__PURE__*/_interopDefaultLegacy(ts);
const prettier__default = /*#__PURE__*/_interopDefaultLegacy(prettier);
const parserBabel__default = /*#__PURE__*/_interopDefaultLegacy(parserBabel);

const RESERVED_IDENTIFIERS = /* @__PURE__ */ new Set(["arguments", "get", "set"]);
function isValidPropName(name) {
  return /^[a-zA-Z_$][a-zA-Z_$0-9]*$/.test(name) && !RESERVED_IDENTIFIERS.has(name);
}

function array(items) {
  return {
    kind: "array",
    items
  };
}
const EMPTY_ARRAY = array([]);
function boolean(value) {
  return {
    kind: "boolean",
    value
  };
}
const FALSE = boolean(false);
const TRUE = boolean(true);
function fn(returnValue) {
  return {
    kind: "function",
    returnValue
  };
}
function map(values) {
  return {
    kind: "map",
    values
  };
}
const NULL = { kind: "null" };
function number(value) {
  return {
    kind: "number",
    value
  };
}
function object(entries) {
  return {
    kind: "object",
    entries
  };
}
function promise(value) {
  return {
    kind: "promise",
    value
  };
}
const EMPTY_OBJECT = object([]);
const EMPTY_MAP = map(EMPTY_OBJECT);
function set(values) {
  return {
    kind: "set",
    values
  };
}
const EMPTY_SET = set(EMPTY_ARRAY);
function string(value) {
  return {
    kind: "string",
    value
  };
}
const UNDEFINED = { kind: "undefined" };
function unknown(source) {
  return {
    kind: "unknown",
    source
  };
}
const UNKNOWN = unknown();

function generateSerializableValue(type, collected, isFunctionReturnValue = false, path = []) {
  return _generateSerializableValue(type, collected, path, [], isFunctionReturnValue);
}
function _generateSerializableValue(type, collected, path, rejectTypeNames, isFunctionReturnValue) {
  let encounteredAliases;
  [type, encounteredAliases] = typeAnalyzer.dereferenceType(type, collected, rejectTypeNames);
  rejectTypeNames = [...rejectTypeNames, ...encounteredAliases];
  switch (type.kind) {
    case "any":
    case "unknown":
    case "never":
    case "void":
      return UNDEFINED;
    case "null":
      return NULL;
    case "boolean":
      return FALSE;
    case "string":
    case "node":
      return string(path.length === 0 ? "node" : path[path.length - 1]);
    case "number":
      return number(0);
    case "literal":
      if (typeof type.value === "number") {
        return number(type.value);
      } else if (typeof type.value === "string") {
        return string(type.value);
      } else {
        return type.value ? TRUE : FALSE;
      }
    case "enum": {
      const value = Object.values(type.options)[0];
      if (typeof value === "number") {
        return number(value);
      } else {
        return string(value || "unknown");
      }
    }
    case "array":
      return generateArrayValue(type, collected, path, rejectTypeNames, isFunctionReturnValue);
    case "set": {
      return set(generateArrayValue(typeAnalyzer.arrayType(type.items), collected, path, rejectTypeNames, isFunctionReturnValue));
    }
    case "object": {
      const entries = [];
      for (const [propName, propType] of Object.entries(type.fields)) {
        const propValue = _generateSerializableValue(propType, collected, [...path, propName], rejectTypeNames, isFunctionReturnValue);
        if (propValue.kind === "undefined") {
          continue;
        }
        if (!isValidPropName(propName)) {
          continue;
        }
        entries.push({
          key: string(propName),
          value: propValue
        });
      }
      return object(entries);
    }
    case "map":
      return map(EMPTY_OBJECT);
    case "record":
      return EMPTY_OBJECT;
    case "union":
      if (typeAnalyzer.isValid(type, collected, void 0)) {
        return UNDEFINED;
      }
      if (typeAnalyzer.isValid(type, collected, null)) {
        return NULL;
      }
      if (typeAnalyzer.isValid(type, collected, false)) {
        return FALSE;
      }
      return _generateSerializableValue(type.types[0], collected, path, rejectTypeNames, isFunctionReturnValue);
    case "intersection":
      return _generateSerializableValue(type.types[0], collected, path, rejectTypeNames, isFunctionReturnValue);
    case "function":
      return fn(isFunctionReturnValue ? UNDEFINED : _generateSerializableValue(type.returnType, collected, path, rejectTypeNames, true));
    case "optional":
      return UNDEFINED;
    case "promise": {
      return promise({
        type: "reject",
        message: null
      });
    }
    case "name":
      return _generateSerializableValue(type, collected, path, rejectTypeNames, isFunctionReturnValue);
    default:
      throw assertNever.assertNever(type);
  }
}
function generateArrayValue(type, collected, path, rejectTypeNames, isFunctionReturnValue) {
  if (isFunctionReturnValue) {
    return EMPTY_ARRAY;
  }
  const itemValue = _generateSerializableValue(type.items, collected, path, rejectTypeNames, isFunctionReturnValue);
  if (itemValue.kind === "undefined" || itemValue.kind === "object" && Object.keys(itemValue.entries).length === 0) {
    return EMPTY_ARRAY;
  }
  return array([itemValue]);
}

function parseSerializableValue(expression) {
  const fallbackValue = unknown(expression.getText());
  if (ts__default.isParenthesizedExpression(expression)) {
    return parseSerializableValue(expression.expression);
  }
  if (expression.kind === ts__default.SyntaxKind.NullKeyword) {
    return NULL;
  }
  if (expression.kind === ts__default.SyntaxKind.TrueKeyword) {
    return TRUE;
  }
  if (expression.kind === ts__default.SyntaxKind.FalseKeyword) {
    return FALSE;
  }
  if (expression.kind === ts__default.SyntaxKind.UndefinedKeyword || ts__default.isIdentifier(expression) && expression.text === "undefined") {
    return UNDEFINED;
  }
  if (ts__default.isNumericLiteral(expression)) {
    try {
      return number(parseFloat(expression.text));
    } catch {
      return fallbackValue;
    }
  }
  if (ts__default.isPrefixUnaryExpression(expression) && expression.operator === ts__default.SyntaxKind.MinusToken && ts__default.isNumericLiteral(expression.operand)) {
    try {
      return number(-parseFloat(expression.operand.text));
    } catch {
      return fallbackValue;
    }
  }
  if (ts__default.isStringLiteral(expression)) {
    return string(expression.text);
  }
  if (ts__default.isNoSubstitutionTemplateLiteral(expression)) {
    return string(expression.text);
  }
  if (ts__default.isArrayLiteralExpression(expression)) {
    return array(expression.elements.map(parseSerializableValue));
  }
  if (ts__default.isNewExpression(expression) && (expression.arguments?.length || 0) <= 1) {
    if (ts__default.isIdentifier(expression.expression)) {
      if (expression.expression.text === "Set") {
        const firstArgument = expression.arguments && expression.arguments[0];
        if (!firstArgument) {
          return EMPTY_SET;
        } else if (ts__default.isArrayLiteralExpression(firstArgument)) {
          return set(array(firstArgument.elements.map(parseSerializableValue)));
        }
      }
      if (expression.expression.text === "Map") {
        const firstArgument = expression.arguments && expression.arguments[0];
        if (!firstArgument) {
          return EMPTY_MAP;
        } else if (ts__default.isArrayLiteralExpression(firstArgument)) {
          const entries = [];
          for (const element of firstArgument.elements) {
            if (!ts__default.isArrayLiteralExpression(element) || element.elements.length !== 2) {
              return fallbackValue;
            }
            entries.push({
              key: parseSerializableValue(element.elements[0]),
              value: parseSerializableValue(element.elements[1])
            });
          }
          return map(object(entries));
        } else if (ts__default.isCallExpression(firstArgument) && ts__default.isPropertyAccessExpression(firstArgument.expression) && ts__default.isIdentifier(firstArgument.expression.expression) && firstArgument.expression.expression.text === "Object" && firstArgument.expression.name.text === "entries" && firstArgument.arguments.length === 1) {
          const object2 = parseSerializableValue(firstArgument.arguments[0]);
          if (object2.kind !== "object") {
            return fallbackValue;
          }
          return map(object2);
        }
      }
    }
  }
  if (ts__default.isCallExpression(expression) && ts__default.isPropertyAccessExpression(expression.expression) && ts__default.isIdentifier(expression.expression.expression) && expression.expression.expression.text === "Promise" && ts__default.isIdentifier(expression.expression.name) && expression.arguments.length <= 1) {
    const firstArgument = expression.arguments[0];
    if (expression.expression.name.text === "resolve") {
      const value = firstArgument ? parseSerializableValue(firstArgument) : UNDEFINED;
      return promise({
        type: "resolve",
        value
      });
    }
    if (expression.expression.name.text === "reject") {
      if (!firstArgument) {
        return promise({
          type: "reject",
          message: null
        });
      } else if (ts__default.isNewExpression(firstArgument) && (firstArgument.arguments?.length || 0) <= 1) {
        const firstErrorArgument = firstArgument.arguments && firstArgument.arguments[0];
        if (firstErrorArgument) {
          if (ts__default.isStringLiteral(firstErrorArgument) || ts__default.isNoSubstitutionTemplateLiteral(firstErrorArgument)) {
            return promise({
              type: "reject",
              message: firstErrorArgument.text
            });
          }
        }
      }
    }
  }
  if (ts__default.isObjectLiteralExpression(expression)) {
    const entries = [];
    for (const property of expression.properties) {
      if (!ts__default.isPropertyAssignment(property)) {
        return fallbackValue;
      }
      if (ts__default.isIdentifier(property.name) || ts__default.isStringLiteral(property.name) || ts__default.isNumericLiteral(property.name)) {
        entries.push({
          key: string(property.name.text),
          value: parseSerializableValue(property.initializer)
        });
      } else {
        return fallbackValue;
      }
    }
    return object(entries);
  }
  if (ts__default.isArrowFunction(expression) && expression.parameters.length === 0) {
    if (ts__default.isBlock(expression.body)) {
      const statements = expression.body.statements;
      if (statements.length === 0) {
        return fn(UNDEFINED);
      } else if (statements.length === 1) {
        const returnStatement = statements[0];
        if (!ts__default.isReturnStatement(returnStatement)) {
          return fallbackValue;
        }
        return fn(returnStatement.expression ? parseSerializableValue(returnStatement.expression) : UNDEFINED);
      } else {
        return fallbackValue;
      }
    } else {
      return fn(parseSerializableValue(expression.body));
    }
  }
  if (ts__default.isFunctionExpression(expression) && expression.parameters.length === 0) {
    const statements = expression.body.statements;
    if (statements.length === 0) {
      return fn(UNDEFINED);
    } else if (statements.length === 1) {
      const returnStatement = statements[0];
      if (!ts__default.isReturnStatement(returnStatement)) {
        return fallbackValue;
      }
      return fn(returnStatement.expression ? parseSerializableValue(returnStatement.expression) : UNDEFINED);
    } else {
      return fallbackValue;
    }
  }
  return fallbackValue;
}

function serializableValueToJavaScript(value) {
  let expression = serializableValueToUnformattedJavaScript(value);
  try {
    expression = formatExpression(expression);
  } catch {
  }
  return expression;
}
function formatExpression(expressionSource) {
  const formattedStatement = prettier__default.format(`value = ${expressionSource}`, {
    parser: "babel",
    plugins: [parserBabel__default],
    filepath: "component.js",
    trailingComma: "none"
  }).trim();
  return formattedStatement.replace(/^value = ((.|\s)*);$/m, "$1");
}
function serializableValueToUnformattedJavaScript(value) {
  switch (value.kind) {
    case "array":
      return `[${value.items.map((item) => serializableValueToUnformattedJavaScript(item)).join(", ")}]`;
    case "boolean":
      return value.value ? "true" : "false";
    case "function":
      return `() => ${value.returnValue.kind === "undefined" ? "{}" : `(${serializableValueToUnformattedJavaScript(value.returnValue)})`}`;
    case "map":
      return `new Map(${value.values.entries.length > 0 ? `Object.entries(${serializableValueToUnformattedJavaScript(value.values)})` : ""})`;
    case "null":
      return "null";
    case "number":
      return value.value.toString(10);
    case "object": {
      if (Object.entries(value.entries).length === 0) {
        return "{}";
      }
      let text = "";
      text += "{\n";
      for (const entry of value.entries) {
        text += `${entry.key.kind === "string" ? JSON.stringify(entry.key.value) : `[${serializableValueToUnformattedJavaScript(entry.key)}]`}: ${serializableValueToUnformattedJavaScript(entry.value)},
`;
      }
      text += "\n}";
      return text;
    }
    case "promise":
      if (value.value.type === "reject") {
        return `Promise.reject(${value.value.message === null ? "" : `new Error(${JSON.stringify(value.value.message)})`})`;
      } else {
        return `Promise.resolve(${serializableValueToUnformattedJavaScript(value.value.value)})`;
      }
    case "set":
      return `new Set(${value.values.items.length > 0 ? serializableValueToUnformattedJavaScript(value.values) : ""})`;
    case "string":
      return JSON.stringify(value.value);
    case "undefined":
      return "undefined";
    case "unknown": {
      const source = value.source ?? "{}";
      try {
        formatExpression(source);
        return source;
      } catch {
        return "{}";
      }
    }
    default:
      throw assertNever__default(value);
  }
}

exports.EMPTY_ARRAY = EMPTY_ARRAY;
exports.EMPTY_MAP = EMPTY_MAP;
exports.EMPTY_OBJECT = EMPTY_OBJECT;
exports.EMPTY_SET = EMPTY_SET;
exports.FALSE = FALSE;
exports.NULL = NULL;
exports.TRUE = TRUE;
exports.UNDEFINED = UNDEFINED;
exports.UNKNOWN = UNKNOWN;
exports.array = array;
exports.boolean = boolean;
exports.fn = fn;
exports.generateSerializableValue = generateSerializableValue;
exports.map = map;
exports.number = number;
exports.object = object;
exports.parseSerializableValue = parseSerializableValue;
exports.promise = promise;
exports.serializableValueToJavaScript = serializableValueToJavaScript;
exports.set = set;
exports.string = string;
exports.unknown = unknown;
